#include <cstddef>
using std::size_t;

#include <cassert>

#include <cmath>
using std::fabs;
using std::fmod;
using std::atan2;

#include <iostream>
using std::cout;
using std::cerr;
using std::endl;

#include <fstream>
using std::ofstream;

#include <iomanip>
using std::setw;

#include <algorithm>
using std::min;
using std::swap;
using std::shuffle;

#include <cmath>
using std::pow;
using std::fabs;

#include <vector>
using std::vector;

#ifdef USE_OPENGL
#ifdef __APPLE__
#include <GLUT/glut.h>
#else
#include <GL/glut.h>
#endif
#endif // USE_OPENGL

#include <random>
using std::mt19937;
using std::uniform_real_distribution;

#include <chrono>
using std::chrono::high_resolution_clock;

#include <functional>
using std::bind;

static mt19937 rng; // mersenne twistor random number generator
static auto rnd = bind(uniform_real_distribution<double>(0,1),rng); // random number in [0,1)

enum vel_t { NONE = 0, N = 1, E = 2, SE = 4, S = 8, W = 16, NW = 32, STATIONARY = 64, OBSTACLE = 128 };

const size_t halfL = 320;
const size_t L = 2*halfL;
const size_t L2 = L*L;

unsigned char buffer1[L2];
unsigned char buffer2[L2];
unsigned char* fluid = buffer1;
unsigned char* fluid_next = buffer2;

const size_t bins = 40;
double px_coarse[bins][bins];
double py_coarse[bins][bins];
unsigned long int num_coarse[bins][bins];

#ifdef USE_OPENGL
const int delay = 0; // milliseconds
#endif // USE_OPENGL

size_t index(size_t i, size_t j) { return i+L*j; }
size_t indexE(size_t i, size_t j) { return index((i+1)%L,j); }
size_t indexW(size_t i, size_t j) { return index((i+L-1)%L,j); }
size_t indexN(size_t i, size_t j) { /* (!) code required */ return 0; }
size_t indexS(size_t i, size_t j) { /* (!) code required */ return 0; }
size_t indexNW(size_t i, size_t j) { /* (!) code required */ return 0; }
size_t indexSE(size_t i, size_t j) { /* (!) code required */ return 0; }

double xpos(size_t i, size_t j) { return fmod(2.0*(i+0.5*j)/L,2.0) - 1.0; }
double ypos(size_t j) { return 0.5*sqrt(3.0)*(fmod(2.0*j/L,2.0) - 1.0); }

unsigned int occupancy(unsigned char v) { /* (!) code required */ return 0; }

unsigned int in_motion(unsigned char v) { return ((v & N ? 1 : 0) + (v & E ? 1 : 0) + (v & SE ? 1 : 0) + (v & S ? 1 : 0) + (v & W ? 1 : 0) + (v & NW ? 1 : 0)); }

bool is_flowing = true;
#ifdef USE_OPENGL
const double d = 2.0/L;
void display()
{
	glClear(GL_COLOR_BUFFER_BIT);
	for (size_t i = 0; i < L; ++i)
		for (size_t j = (is_flowing ? 4 : 0); j < L; ++j)
		{
			unsigned char v = fluid[index(i,j)];
			if (v!=0)
			{
				if (v & OBSTACLE)
					glColor3f(1.0, 0.5, 0.25); 
				else
				{
					double shade = 0.5+occupancy(v)/14.0; 
					glColor3f(shade,shade,shade);				
				}
				const double x = xpos(i,j);
				const double y = ypos(j);
				glBegin(GL_POLYGON);
				glVertex2f(x,y);
				glVertex2f(x+d,y);
				glVertex2f(x+d,y+d);
				glVertex2f(x,y+d);
				glEnd();
			}
		}
	glFlush();
	glutSwapBuffers();
}

void keyboard(unsigned char key, int x, int y)
{
	if (key == 'q')
		exit(0);
}
#endif // USE_OPENGL

void inc_momentum(double &px, double &py, unsigned char v)
{
	/* (!) code required */ return;
}

double concentration;
int setup;
void initialize_grid(void)
{
	switch(setup)
	{
		case 0:
			is_flowing = false;
			for (size_t i = 0; i < L; ++i)
				for (size_t j = 0; j < L; ++j)
					fluid[index(i,j)] = ( (rnd() < concentration ? 1 : 0)*N  + 
										 (rnd() < concentration ? 1 : 0)*E  + 
										 (rnd() < concentration ? 1 : 0)*SE + 
										 (rnd() < concentration ? 1 : 0)*S  + 
										 (rnd() < concentration ? 1 : 0)*W  + 
										 (rnd() < concentration ? 1 : 0)*NW +
										 (rnd() < concentration ? 1 : 0)*STATIONARY );
			return;
		case 1:
			is_flowing = false;	
			for (size_t i = 0; i < L; ++i)
				for (size_t j = 0; j < L; ++j)
				{
					const double dx = xpos(i,j);
					const double dy = ypos(j);
					fluid[index(i,j)] = ( dx*dx + dy*dy < 0.03 ? 0 : ( (rnd() < concentration ? 1 : 0)*N  + 
																	  (rnd() < concentration ? 1 : 0)*E  + 
																	  (rnd() < concentration ? 1 : 0)*SE + 
																	  (rnd() < concentration ? 1 : 0)*S  + 
																	  (rnd() < concentration ? 1 : 0)*W  + 
																	  (rnd() < concentration ? 1 : 0)*NW ) );
				}
			return;
		case 2:
			is_flowing = false;	
			for (size_t i = 0; i < L; ++i)
				for (size_t j = 0; j < L; ++j)
				{
					const double dx = xpos(i,j);
					fluid[index(i,j)] = ( dx*dx  < 0.001 ? 0 : ( (rnd() < concentration ? 1 : 0)*N  + 
																(rnd() < concentration ? 1 : 0)*E  + 
																(rnd() < concentration ? 1 : 0)*SE + 
																(rnd() < concentration ? 1 : 0)*S  + 
																(rnd() < concentration ? 1 : 0)*W  + 
																(rnd() < concentration ? 1 : 0)*NW ) );
				}
			return;
		case 3:
			is_flowing = false;	
			for (size_t i = 0; i < L; ++i)
				for (size_t j = 0; j < L; ++j)
				{
					const double dx = xpos(i,j);
					const double dy = ypos(j);
					const double len = pow(fabs(dx),1.4) + pow(fabs(dy),1.4);
					if (dx < -0.98 or dx > 0.98 or dy < -0.85 or dy > 0.85)
						fluid[index(i,j)] = OBSTACLE;
					else if (len < 0.32)
						fluid[index(i,j)] =   (rnd() < concentration ? 1 : 0)*N  + 
						(rnd() < concentration ? 1 : 0)*E  + 
						(rnd() < concentration ? 1 : 0)*SE + 
						(rnd() < concentration ? 1 : 0)*S  + 
						(rnd() < concentration ? 1 : 0)*W  + 
						(rnd() < concentration ? 1 : 0)*NW;
					else if (0.32 <= len and len < 0.35 and !(fabs(dx) < 0.03 and dy > 0.0))
						fluid[index(i,j)] = OBSTACLE;
					else
						fluid[index(i,j)] =   (rnd() < 0.1*concentration ? 1 : 0)*N  + 
						(rnd() < 0.1*concentration ? 1 : 0)*E  + 
						(rnd() < 0.1*concentration ? 1 : 0)*SE + 
						(rnd() < 0.1*concentration ? 1 : 0)*S  + 
						(rnd() < 0.1*concentration ? 1 : 0)*W  + 
						(rnd() < 0.1*concentration ? 1 : 0)*NW;
				}
			return;
		case 4:
			is_flowing = true;
			for (size_t i = 0; i < L; ++i)
				for (size_t j = 0; j < L; ++j)
				{
					const double dx = xpos(i,j);
					const double dy = ypos(j)+0.15;
					const double len = sqrt(2.0*dx*dx + dy*dy);
					if (len < 0.25)
						fluid[index(i,j)] = (len < 0.22 ? NONE : OBSTACLE);
					else
						fluid[index(i,j)] = (rnd() < concentration ? 1 : 0)*N  + 
						(rnd() < concentration ? 1 : 0)*E  + 
						(rnd() < concentration ? 1 : 0)*SE + 
						(rnd() < concentration ? 1 : 0)*S  + 
						(rnd() < concentration ? 1 : 0)*W  + 
						(rnd() < concentration ? 1 : 0)*NW;
				}
			return;
	}
}

void propagate(void)
{
	for (size_t i = 0; i < L; ++i)
		for (size_t j = 0; j < L; ++j)
			fluid_next[index(i,j)] = (fluid[indexS(i,j)] & N) | (fluid[indexW(i,j)] & E) | (fluid[indexNW(i,j)] & SE) |
			(fluid[indexN(i,j)] & S) | (fluid[indexE(i,j)] & W) | (fluid[indexSE(i,j)] & NW) | 
			(fluid[index(i,j)] & STATIONARY) | (fluid[index(i,j)] & OBSTACLE);
	
	swap(fluid,fluid_next);
}

void collision(void)
{
	/*
	 //      N
	 //  NW
	 //
	 //   W      E
	 //
	 //      S  SE
	 */
	for (size_t i = 0; i < L; ++i)
		for (size_t j = 0; j < L; ++j)
		{
			unsigned char& v = fluid[index(i,j)];
			const unsigned int n = in_motion(v);
			if (v & OBSTACLE)
			{
				static vector<size_t> unobstructed(6);
				unobstructed.resize(0);
				if (!(fluid[indexN(i,j)] & OBSTACLE)) unobstructed.push_back(N);
				if (!(fluid[indexE(i,j)] & OBSTACLE)) unobstructed.push_back(E);
				if (!(fluid[indexS(i,j)] & OBSTACLE)) unobstructed.push_back(S);
				if (!(fluid[indexSE(i,j)] & OBSTACLE)) unobstructed.push_back(SE);
				if (!(fluid[indexW(i,j)] & OBSTACLE)) unobstructed.push_back(W);
				if (!(fluid[indexNW(i,j)] & OBSTACLE)) unobstructed.push_back(NW);
				shuffle(unobstructed.begin(),unobstructed.end(),rng);
				const unsigned char vold = v;
				v = OBSTACLE;
				if (vold & N) { v |= unobstructed.back(); unobstructed.pop_back(); }
				if (vold & E) { v |= unobstructed.back(); unobstructed.pop_back(); }
				if (vold & SE) { v |= unobstructed.back(); unobstructed.pop_back(); }
				if (vold & S) { v |= unobstructed.back(); unobstructed.pop_back(); }
				if (vold & W) { v |= unobstructed.back(); unobstructed.pop_back(); }
				if (vold & NW) { v |= unobstructed.back(); unobstructed.pop_back(); }
			}
			else if (n == 1 and (v & STATIONARY))
			{
				if (v & E) v ^= N | E | SE | STATIONARY;
				else if (v & SE) v ^= E | SE | S | STATIONARY;
				else if (v & S) v ^= SE | S | W | STATIONARY;
				else if (v & W) v ^= S | W | NW | STATIONARY;
				else if (v & NW) v ^= W | NW | N | STATIONARY;
				else if (v & N) v ^= NW | N | E | STATIONARY;
			}		 
			else if (n == 2)
			{	
				if ( !(v & STATIONARY) and (v & N) and (v & SE) ) v ^= N | E | SE | STATIONARY;
				else if ( !(v & STATIONARY) and (v & E) and (v & S) ) v ^= E | SE | S | STATIONARY;
				else if ( !(v & STATIONARY) and (v & SE) and (v & W) ) v ^= SE | S | W | STATIONARY;
				else if ( !(v & STATIONARY) and (v & S) and (v & NW) ) v ^= S | W | NW | STATIONARY;
				else if ( !(v & STATIONARY) and (v & W) and (v & N) ) v ^= W | NW | N | STATIONARY;
				else if ( !(v & STATIONARY) and (v & NW) and (v & E) ) v ^= NW | N | E | STATIONARY;
				else if ( (v & N) and (v & S) ) v ^= N | S | ( rnd() < 0.5 ? (W | E) : (NW | SE) );
				else if ( (v & W) and (v & E) ) v ^= W | E | ( rnd() < 0.5 ? (N | S) : (NW | SE) );
				else if ( (v & NW) and (v & SE) ) v ^= NW | SE | ( rnd() < 0.5 ? (W | E) : (N | S) );
			}
			else if (n == 3)
			{
				
				if ( (v & N) and (v & SE) and (v & W) ) v ^= N | SE | W | NW | E | S;
				else if ( (v & NW) and (v & E) and (v & S) ) v ^= NW | E | S | N | SE | W;
				
				else if ( (v & NW) and (v & SE) and (v & S) ) v ^= NW | SE | W | E;
				else if ( (v & SE) and (v & S) and (v & N) ) v ^= S | N | W | E;
				
				else if ( (v & N) and (v & S) and (v & W) ) v ^= N | S | NW | SE;
				else if ( (v & S) and (v & W) and (v & E) ) v ^= W | E | NW | SE;
				
				else if ( (v & E) and (v & W) and (v & NW) ) v ^= E | W | N | S;
				else if ( (v & W) and (v & NW) and (v & SE) ) v ^= NW | SE | N | S;
				
				else if ( (v & SE) and (v & NW) and (v & N) ) v ^= SE | NW | E | W;
				else if ( (v & NW) and (v & N) and (v & S) ) v ^= N | S | E | W;
				
				else if ( (v & S) and (v & N) and (v & E) ) v ^= S | N | SE | NW;
				else if ( (v & N) and (v & E) and (v & W) ) v ^= E | W | SE | NW;
				
				else if ( (v & W) and (v & E) and (v & SE) ) v ^= W | E | S | N;
				else if ( (v & E) and (v & SE) and (v & NW) ) v ^= SE | NW | S | N;
			}
			/* (!) code required */
		}
}

void flow(void)
{
	for (size_t i = 0; i < L; ++i)
	{
		fluid[index(i,L-1)] = NONE; 
		fluid[index(i,0)] = NONE;
		if (rnd() < concentration) fluid[index(i,0)] |= NW;
		if (rnd() < concentration) fluid[index(i,0)] |= N;
		if (rnd() < concentration) fluid[index(i,1)] |= NW;
		if (rnd() < concentration) fluid[index(i,1)] |= N;
		if (rnd() < concentration) fluid[index(i,1)] |= W;
		if (rnd() < concentration) fluid[index(i,1)] |= E;
		if (rnd() < concentration) fluid[index(i,2)] |= NW;
		if (rnd() < concentration) fluid[index(i,2)] |= N;
		if (rnd() < concentration) fluid[index(i,2)] |= W;
		if (rnd() < concentration) fluid[index(i,2)] |= E;		
		if (rnd() < concentration) fluid[index(i,3)] |= NW;
		if (rnd() < concentration) fluid[index(i,3)] |= N;
		if (rnd() < concentration) fluid[index(i,3)] |= W;
		if (rnd() < concentration) fluid[index(i,3)] |= E;		
	}
}

void report(void)
{
	unsigned long int n = 0;
	double px = 0.0;
	double py = 0.0;
	for (size_t i = 0; i < L; ++i)
		for (size_t j = 0; j < L; ++j)
		{
			unsigned char& v = fluid[index(i,j)];
			n += occupancy(v);
			inc_momentum(px,py,v);
		}
	cout << setw(15) << n << setw(15) << px << setw(15) << py << endl;
}

void animate(int)
{
	propagate();
	collision();
	if (is_flowing) flow();
	
	report();
	
#ifdef USE_OPENGL
	glutPostRedisplay();
	glutTimerFunc(delay,animate,0);
#endif // USE_OPENGL
	
}

int main(int argc, char **argv)
{
#ifdef USE_OPENGL
	glutInit(&argc,argv);
#endif // USE_OPENGL
	
	if (argc != 3)
	{
#ifdef USE_OPENGL
		cerr << "Usage: lattice_gas_openGL (setup=0,1,..,4) (0 < concentration < 1)" << endl;
#else
		cerr << "Usage: lattice_gas (setup=0,1,..,4) (0 < concentration < 1)" << endl;		
#endif
		exit(1);
	}
	setup = atoi(argv[1]);
	concentration = atof(argv[2]);
	if (setup < 0 or setup > 5) 
	{
		cerr << "Error: `setup` parameter must be 0, 1, 2, 3, or 4" << endl;
		exit(0);
	}
	if (concentration < 0.0 or concentration > 1.0) 
	{
		cerr << "Error: `concentration` parameter must be in the interval [0,1]" << endl;
		exit(0);
	}
   
   rng.seed(mt19937::result_type(high_resolution_clock::now().time_since_epoch().count()));
   
	initialize_grid();	
	
#ifdef USE_OPENGL
	glutInitDisplayMode (GLUT_RGB | GLUT_DOUBLE);
	glutInitWindowSize (640, 640);
	glutInitWindowPosition (100, 100);
	glutCreateWindow("FHP model lattice hydrodynamics");
	glutDisplayFunc(display);
	glutKeyboardFunc(keyboard);	
	glutTimerFunc(delay,animate,0);
	glutMainLoop();
#else
	while (true) animate(0);
#endif // USE_OPENGL	
}
